# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Schemas were moved to mlrun.common.schemas.
For backwards compatibility reasons, we have left this folder untouched so that older version of mlrun would
   be able to migrate to newer version without having to upgrade into an intermediate version.

The DeprecationHelper class is used to print a deprecation warning when the old import is used, and return the new
schema.
"""

import sys
import warnings

import mlrun.common.formatters
import mlrun.common.schemas
import mlrun.common.schemas.artifact as old_artifact
import mlrun.common.schemas.auth as old_auth
import mlrun.common.schemas.background_task as old_background_task
import mlrun.common.schemas.client_spec as old_client_spec
import mlrun.common.schemas.clusterization_spec as old_clusterization_spec
import mlrun.common.schemas.constants as old_constants
import mlrun.common.schemas.feature_store as old_feature_store
import mlrun.common.schemas.frontend_spec as old_frontend_spec
import mlrun.common.schemas.function as old_function
import mlrun.common.schemas.http as old_http
import mlrun.common.schemas.k8s as old_k8s
import mlrun.common.schemas.memory_reports as old_memory_reports
import mlrun.common.schemas.model_monitoring.grafana
import mlrun.common.schemas.object as old_object
import mlrun.common.schemas.pipeline as old_pipeline
import mlrun.common.schemas.project as old_project
import mlrun.common.schemas.runtime_resource as old_runtime_resource
import mlrun.common.schemas.schedule as old_schedule
import mlrun.common.schemas.secret as old_secret
import mlrun.common.schemas.tag as old_tag


class DeprecationHelper:
    """A helper class to deprecate old schemas"""

    def __init__(self, new_target, version="1.4.0"):
        self._new_target = new_target
        self._version = version

    def _warn(self):
        warnings.warn(
            f"mlrun.api.schemas.{self._new_target.__name__} is deprecated since version {self._version}, "
            f"Use mlrun.common.schemas.{self._new_target.__name__} instead.",
            FutureWarning,
        )

    def __call__(self, *args, **kwargs):
        self._warn()
        return self._new_target(*args, **kwargs)

    def __getattr__(self, attr):
        self._warn()
        return getattr(self._new_target, attr)


# for backwards compatibility, we need to inject the old import path to `sys.modules`
sys.modules["mlrun.api.schemas.artifact"] = old_artifact
sys.modules["mlrun.api.schemas.auth"] = old_auth
sys.modules["mlrun.api.schemas.background_task"] = old_background_task
sys.modules["mlrun.api.schemas.client_spec"] = old_client_spec
sys.modules["mlrun.api.schemas.clusterization_spec"] = old_clusterization_spec
sys.modules["mlrun.api.schemas.constants"] = old_constants
sys.modules["mlrun.api.schemas.feature_store"] = old_feature_store
sys.modules["mlrun.api.schemas.frontend_spec"] = old_frontend_spec
sys.modules["mlrun.api.schemas.function"] = old_function
sys.modules["mlrun.api.schemas.http"] = old_http
sys.modules["mlrun.api.schemas.k8s"] = old_k8s
sys.modules["mlrun.api.schemas.memory_reports"] = old_memory_reports
sys.modules["mlrun.api.schemas.object"] = old_object
sys.modules["mlrun.api.schemas.pipeline"] = old_pipeline
sys.modules["mlrun.api.schemas.project"] = old_project
sys.modules["mlrun.api.schemas.runtime_resource"] = old_runtime_resource
sys.modules["mlrun.api.schemas.schedule"] = old_schedule
sys.modules["mlrun.api.schemas.secret"] = old_secret
sys.modules["mlrun.api.schemas.tag"] = old_tag

# The DeprecationHelper class is used to print a deprecation warning when the old import is used,
# and return the new schema. This is done for backwards compatibility with mlrun.api.schemas.
ArtifactCategories = DeprecationHelper(mlrun.common.schemas.ArtifactCategories)
ArtifactIdentifier = DeprecationHelper(mlrun.common.schemas.ArtifactIdentifier)
ArtifactsFormat = DeprecationHelper(mlrun.common.formatters.ArtifactFormat)
AuthInfo = DeprecationHelper(mlrun.common.schemas.AuthInfo)
AuthorizationAction = DeprecationHelper(mlrun.common.schemas.AuthorizationAction)
AuthorizationResourceTypes = DeprecationHelper(
    mlrun.common.schemas.AuthorizationResourceTypes
)
AuthorizationVerificationInput = DeprecationHelper(
    mlrun.common.schemas.AuthorizationVerificationInput
)
Credentials = DeprecationHelper(mlrun.common.schemas.Credentials)
ProjectsRole = DeprecationHelper(mlrun.common.schemas.ProjectsRole)

BackgroundTask = DeprecationHelper(mlrun.common.schemas.BackgroundTask)
BackgroundTaskMetadata = DeprecationHelper(mlrun.common.schemas.BackgroundTaskMetadata)
BackgroundTaskSpec = DeprecationHelper(mlrun.common.schemas.BackgroundTaskSpec)
BackgroundTaskState = DeprecationHelper(mlrun.common.schemas.BackgroundTaskState)
BackgroundTaskStatus = DeprecationHelper(mlrun.common.schemas.BackgroundTaskStatus)
ClientSpe = DeprecationHelper(mlrun.common.schemas.ClientSpec)
ClusterizationSpec = DeprecationHelper(mlrun.common.schemas.ClusterizationSpec)
WaitForChiefToReachOnlineStateFeatureFlag = DeprecationHelper(
    mlrun.common.schemas.WaitForChiefToReachOnlineStateFeatureFlag
)
APIStates = DeprecationHelper(mlrun.common.schemas.APIStates)
ClusterizationRole = DeprecationHelper(mlrun.common.schemas.ClusterizationRole)
DeletionStrategy = DeprecationHelper(mlrun.common.schemas.DeletionStrategy)
FeatureStorePartitionByField = DeprecationHelper(
    mlrun.common.schemas.FeatureStorePartitionByField
)
HeaderNames = DeprecationHelper(mlrun.common.schemas.HeaderNames)
LogsCollectorMode = DeprecationHelper(mlrun.common.schemas.LogsCollectorMode)
OrderType = DeprecationHelper(mlrun.common.schemas.OrderType)
PatchMode = DeprecationHelper(mlrun.common.schemas.PatchMode)
RunPartitionByField = DeprecationHelper(mlrun.common.schemas.RunPartitionByField)
SortField = DeprecationHelper(mlrun.common.schemas.SortField)
EntitiesOutput = DeprecationHelper(mlrun.common.schemas.EntitiesOutput)
Entity = DeprecationHelper(mlrun.common.schemas.Entity)
EntityListOutput = DeprecationHelper(mlrun.common.schemas.EntityListOutput)
EntityRecord = DeprecationHelper(mlrun.common.schemas.EntityRecord)
Feature = DeprecationHelper(mlrun.common.schemas.Feature)
FeatureListOutput = DeprecationHelper(mlrun.common.schemas.FeatureListOutput)
FeatureRecord = DeprecationHelper(mlrun.common.schemas.FeatureRecord)
FeatureSet = DeprecationHelper(mlrun.common.schemas.FeatureSet)
FeatureSetDigestOutput = DeprecationHelper(mlrun.common.schemas.FeatureSetDigestOutput)
FeatureSetDigestSpec = DeprecationHelper(mlrun.common.schemas.FeatureSetDigestSpec)
FeatureSetIngestInput = DeprecationHelper(mlrun.common.schemas.FeatureSetIngestInput)
FeatureSetIngestOutput = DeprecationHelper(mlrun.common.schemas.FeatureSetIngestOutput)
FeatureSetRecord = DeprecationHelper(mlrun.common.schemas.FeatureSetRecord)
FeatureSetsOutput = DeprecationHelper(mlrun.common.schemas.FeatureSetsOutput)
FeatureSetSpec = DeprecationHelper(mlrun.common.schemas.FeatureSetSpec)
FeatureSetsTagsOutput = DeprecationHelper(mlrun.common.schemas.FeatureSetsTagsOutput)
FeaturesOutput = DeprecationHelper(mlrun.common.schemas.FeaturesOutput)
FeatureVector = DeprecationHelper(mlrun.common.schemas.FeatureVector)
FeatureVectorRecord = DeprecationHelper(mlrun.common.schemas.FeatureVectorRecord)
FeatureVectorsOutput = DeprecationHelper(mlrun.common.schemas.FeatureVectorsOutput)
FeatureVectorsTagsOutput = DeprecationHelper(
    mlrun.common.schemas.FeatureVectorsTagsOutput
)
AuthenticationFeatureFlag = DeprecationHelper(
    mlrun.common.schemas.AuthenticationFeatureFlag
)
FeatureFlags = DeprecationHelper(mlrun.common.schemas.FeatureFlags)
FrontendSpec = DeprecationHelper(mlrun.common.schemas.FrontendSpec)
NuclioStreamsFeatureFlag = DeprecationHelper(
    mlrun.common.schemas.NuclioStreamsFeatureFlag
)
PreemptionNodesFeatureFlag = DeprecationHelper(
    mlrun.common.schemas.PreemptionNodesFeatureFlag
)
ProjectMembershipFeatureFlag = DeprecationHelper(
    mlrun.common.schemas.ProjectMembershipFeatureFlag
)
FunctionState = DeprecationHelper(mlrun.common.schemas.FunctionState)
PreemptionModes = DeprecationHelper(mlrun.common.schemas.PreemptionModes)
SecurityContextEnrichmentModes = DeprecationHelper(
    mlrun.common.schemas.SecurityContextEnrichmentModes
)
HTTPSessionRetryMode = DeprecationHelper(mlrun.common.schemas.HTTPSessionRetryMode)
NodeSelectorOperator = DeprecationHelper(mlrun.common.schemas.NodeSelectorOperator)
Resources = DeprecationHelper(mlrun.common.schemas.Resources)
ResourceSpec = DeprecationHelper(mlrun.common.schemas.ResourceSpec)
IndexedHubSource = DeprecationHelper(mlrun.common.schemas.IndexedHubSource)
HubCatalog = DeprecationHelper(mlrun.common.schemas.HubCatalog)
HubItem = DeprecationHelper(mlrun.common.schemas.HubItem)
HubObjectMetadata = DeprecationHelper(mlrun.common.schemas.HubObjectMetadata)
HubSource = DeprecationHelper(mlrun.common.schemas.HubSource)
HubSourceSpec = DeprecationHelper(mlrun.common.schemas.HubSourceSpec)
last_source_index = DeprecationHelper(mlrun.common.schemas.last_source_index)
MostCommonObjectTypesReport = DeprecationHelper(
    mlrun.common.schemas.MostCommonObjectTypesReport
)
ObjectTypeReport = DeprecationHelper(mlrun.common.schemas.ObjectTypeReport)
Features = DeprecationHelper(mlrun.common.schemas.Features)
FeatureValues = DeprecationHelper(mlrun.common.schemas.FeatureValues)
GrafanaColumn = DeprecationHelper(
    mlrun.common.schemas.model_monitoring.grafana.GrafanaColumn
)

GrafanaNumberColumn = DeprecationHelper(
    mlrun.common.schemas.model_monitoring.grafana.GrafanaNumberColumn
)
GrafanaStringColumn = DeprecationHelper(
    mlrun.common.schemas.model_monitoring.grafana.GrafanaStringColumn
)
GrafanaTable = DeprecationHelper(
    mlrun.common.schemas.model_monitoring.grafana.GrafanaTable
)
ModelEndpoint = DeprecationHelper(mlrun.common.schemas.ModelEndpoint)
ModelEndpointList = DeprecationHelper(mlrun.common.schemas.ModelEndpointList)
ModelEndpointMetadata = DeprecationHelper(mlrun.common.schemas.ModelEndpointMetadata)
ModelEndpointSpec = DeprecationHelper(mlrun.common.schemas.ModelEndpointSpec)
ModelEndpointStatus = DeprecationHelper(mlrun.common.schemas.ModelEndpointStatus)
NotificationSeverity = DeprecationHelper(mlrun.common.schemas.NotificationSeverity)
NotificationStatus = DeprecationHelper(mlrun.common.schemas.NotificationStatus)
ObjectKind = DeprecationHelper(mlrun.common.schemas.ObjectKind)
ObjectMetadata = DeprecationHelper(mlrun.common.schemas.ObjectMetadata)
ObjectSpec = DeprecationHelper(mlrun.common.schemas.ObjectSpec)
ObjectStatus = DeprecationHelper(mlrun.common.schemas.ObjectStatus)
PipelinesFormat = DeprecationHelper(mlrun.common.formatters.PipelineFormat)
PipelinesOutput = DeprecationHelper(mlrun.common.schemas.PipelinesOutput)
PipelinesPagination = DeprecationHelper(mlrun.common.schemas.PipelinesPagination)
IguazioProject = DeprecationHelper(mlrun.common.schemas.IguazioProject)
Project = DeprecationHelper(mlrun.common.schemas.Project)
ProjectDesiredState = DeprecationHelper(mlrun.common.schemas.ProjectDesiredState)
ProjectMetadata = DeprecationHelper(mlrun.common.schemas.ProjectMetadata)
ProjectOwner = DeprecationHelper(mlrun.common.schemas.ProjectOwner)
ProjectsFormat = DeprecationHelper(mlrun.common.formatters.ProjectFormat)
ProjectsOutput = DeprecationHelper(mlrun.common.schemas.ProjectsOutput)
ProjectSpec = DeprecationHelper(mlrun.common.schemas.ProjectSpec)
ProjectState = DeprecationHelper(mlrun.common.schemas.ProjectState)
ProjectStatus = DeprecationHelper(mlrun.common.schemas.ProjectStatus)
ProjectSummariesOutput = DeprecationHelper(mlrun.common.schemas.ProjectSummariesOutput)
ProjectSummary = DeprecationHelper(mlrun.common.schemas.ProjectSummary)
GroupedByJobRuntimeResourcesOutput = DeprecationHelper(
    mlrun.common.schemas.GroupedByJobRuntimeResourcesOutput
)
GroupedByProjectRuntimeResourcesOutput = DeprecationHelper(
    mlrun.common.schemas.GroupedByProjectRuntimeResourcesOutput
)
KindRuntimeResources = DeprecationHelper(mlrun.common.schemas.KindRuntimeResources)
ListRuntimeResourcesGroupByField = DeprecationHelper(
    mlrun.common.schemas.ListRuntimeResourcesGroupByField
)
RuntimeResource = DeprecationHelper(mlrun.common.schemas.RuntimeResource)
RuntimeResources = DeprecationHelper(mlrun.common.schemas.RuntimeResources)
RuntimeResourcesOutput = DeprecationHelper(mlrun.common.schemas.RuntimeResourcesOutput)
ScheduleCronTrigger = DeprecationHelper(mlrun.common.schemas.ScheduleCronTrigger)
ScheduleInput = DeprecationHelper(mlrun.common.schemas.ScheduleInput)
ScheduleKinds = DeprecationHelper(mlrun.common.schemas.ScheduleKinds)
ScheduleOutput = DeprecationHelper(mlrun.common.schemas.ScheduleOutput)
ScheduleRecord = DeprecationHelper(mlrun.common.schemas.ScheduleRecord)
SchedulesOutput = DeprecationHelper(mlrun.common.schemas.SchedulesOutput)
ScheduleUpdate = DeprecationHelper(mlrun.common.schemas.ScheduleUpdate)
AuthSecretData = DeprecationHelper(mlrun.common.schemas.AuthSecretData)
SecretKeysData = DeprecationHelper(mlrun.common.schemas.SecretKeysData)
SecretProviderName = DeprecationHelper(mlrun.common.schemas.SecretProviderName)
SecretsData = DeprecationHelper(mlrun.common.schemas.SecretsData)
UserSecretCreationRequest = DeprecationHelper(
    mlrun.common.schemas.UserSecretCreationRequest
)
Tag = DeprecationHelper(mlrun.common.schemas.Tag)
TagObjects = DeprecationHelper(mlrun.common.schemas.TagObjects)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from abc import ABC, abstractmethod
from typing import Literal, Optional, Union

from deprecated import deprecated

import mlrun.alerts
import mlrun.common
import mlrun.common.formatters
import mlrun.common.runtimes.constants
import mlrun.common.schemas
import mlrun.common.schemas.model_monitoring.constants as mm_constants
import mlrun.common.schemas.model_monitoring.model_endpoints as mm_endpoints
import mlrun.model_monitoring


class RunDBError(Exception):
    pass


class RunDBInterface(ABC):
    kind = ""

    @abstractmethod
    def connect(self, secrets=None):
        return self

    @abstractmethod
    def store_log(self, uid, project="", body=None, append=False):
        pass

    @abstractmethod
    def get_log(self, uid, project="", offset=0, size=0, attempt=None):
        pass

    @abstractmethod
    def store_run(self, struct, uid, project="", iter=0):
        pass

    @abstractmethod
    def update_run(self, updates: dict, uid, project="", iter=0):
        pass

    @abstractmethod
    def set_run_retrying_status(
        self, project: str, name: str, run_id: str, retrying: bool
    ):
        pass

    @abstractmethod
    def abort_run(self, uid, project="", iter=0, timeout=45, status_text=""):
        pass

    @abstractmethod
    def push_run_notifications(
        self,
        uid,
        project="",
        timeout=45,
    ):
        pass

    def refresh_smtp_configuration(self):
        pass

    def push_pipeline_notifications(
        self,
        pipeline_id,
        project="",
        notifications=None,
        timeout=45,
    ):
        pass

    @abstractmethod
    def read_run(
        self,
        uid: str,
        project: str = "",
        iter: int = 0,
        format_: mlrun.common.formatters.RunFormat = mlrun.common.formatters.RunFormat.full,
    ):
        pass

    @abstractmethod
    def list_runs(
        self,
        name: Optional[str] = None,
        uid: Optional[Union[str, list[str]]] = None,
        project: Optional[str] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        states: Optional[list[mlrun.common.runtimes.constants.RunStates]] = None,
        sort: bool = True,
        iter: bool = False,
        start_time_from: Optional[datetime.datetime] = None,
        start_time_to: Optional[datetime.datetime] = None,
        last_update_time_from: Optional[datetime.datetime] = None,
        last_update_time_to: Optional[datetime.datetime] = None,
        end_time_from: Optional[datetime.datetime] = None,
        end_time_to: Optional[datetime.datetime] = None,
        partition_by: Union[mlrun.common.schemas.RunPartitionByField, str] = None,
        rows_per_partition: int = 1,
        partition_sort_by: Union[mlrun.common.schemas.SortField, str] = None,
        partition_order: Union[
            mlrun.common.schemas.OrderType, str
        ] = mlrun.common.schemas.OrderType.desc,
        max_partitions: int = 0,
        with_notifications: bool = False,
    ):
        pass

    @abstractmethod
    def paginated_list_runs(
        self,
        *args,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
        **kwargs,
    ):
        pass

    @abstractmethod
    def del_run(self, uid, project="", iter=0):
        pass

    @abstractmethod
    def del_runs(
        self,
        name: str = "",
        project: str = "",
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        state: Optional[mlrun.common.runtimes.constants.RunStates] = None,
        days_ago: int = 0,
    ):
        pass

    @abstractmethod
    def store_artifact(
        self,
        key,
        artifact,
        iter=None,
        tag="",
        project="",
        tree=None,
    ):
        pass

    @abstractmethod
    def read_artifact(
        self,
        key,
        tag="",
        iter=None,
        project="",
        tree=None,
        uid=None,
        format_: mlrun.common.formatters.ArtifactFormat = mlrun.common.formatters.ArtifactFormat.full,
    ):
        pass

    @abstractmethod
    def list_artifacts(
        self,
        name: Optional[str] = "",
        project: Optional[str] = "",
        tag: Optional[str] = "",
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        since=None,
        until=None,
        iter: Optional[int] = None,
        best_iteration: bool = False,
        kind: Optional[str] = None,
        category: Union[str, mlrun.common.schemas.ArtifactCategories] = None,
        tree: Optional[str] = None,
        parent: Optional[str] = None,
        format_: mlrun.common.formatters.ArtifactFormat = mlrun.common.formatters.ArtifactFormat.full,
        limit: Optional[int] = None,
        partition_by: Optional[
            Union[mlrun.common.schemas.ArtifactPartitionByField, str]
        ] = None,
        rows_per_partition: int = 1,
        partition_sort_by: Optional[
            Union[mlrun.common.schemas.SortField, str]
        ] = mlrun.common.schemas.SortField.updated,
        partition_order: Union[
            mlrun.common.schemas.OrderType, str
        ] = mlrun.common.schemas.OrderType.desc,
    ):
        pass

    @abstractmethod
    def paginated_list_artifacts(
        self,
        *args,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
        **kwargs,
    ):
        pass

    @abstractmethod
    def del_artifact(
        self,
        key,
        tag="",
        project="",
        tree=None,
        uid=None,
        deletion_strategy: mlrun.common.schemas.artifact.ArtifactsDeletionStrategies = (
            mlrun.common.schemas.artifact.ArtifactsDeletionStrategies.metadata_only
        ),
        secrets: Optional[dict] = None,
        iter=None,
    ):
        pass

    @abstractmethod
    def del_artifacts(
        self,
        name: Optional[str] = "",
        project: Optional[str] = "",
        tag: Optional[str] = "",
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
    ):
        pass

    @abstractmethod
    def store_function(self, function, name, project="", tag="", versioned=False):
        pass

    @abstractmethod
    def get_function(self, name, project="", tag="", hash_key=""):
        pass

    @abstractmethod
    def delete_function(self, name: str, project: str = ""):
        pass

    @abstractmethod
    def list_functions(
        self,
        name: Optional[str] = None,
        project: Optional[str] = None,
        tag: Optional[str] = None,
        kind: Optional[str] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        states: Optional[list[mlrun.common.schemas.FunctionState]] = None,
        format_: mlrun.common.formatters.FunctionFormat = mlrun.common.formatters.FunctionFormat.full,
        since: Optional[datetime.datetime] = None,
        until: Optional[datetime.datetime] = None,
    ):
        pass

    @abstractmethod
    def paginated_list_functions(
        self,
        *args,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
        **kwargs,
    ):
        pass

    @abstractmethod
    def tag_objects(
        self,
        project: str,
        tag_name: str,
        tag_objects: mlrun.common.schemas.TagObjects,
        replace: bool = False,
    ):
        pass

    @abstractmethod
    def delete_objects_tag(
        self,
        project: str,
        tag_name: str,
        tag_objects: mlrun.common.schemas.TagObjects,
    ):
        pass

    @abstractmethod
    def tag_artifacts(
        self,
        artifacts,
        project: str,
        tag_name: str,
        replace: bool = False,
    ):
        pass

    @abstractmethod
    def delete_artifacts_tags(
        self,
        artifacts,
        project: str,
        tag_name: str,
    ):
        pass

    @staticmethod
    def _resolve_artifacts_to_tag_objects(
        artifacts,
    ) -> mlrun.common.schemas.TagObjects:
        """
        :param artifacts: Can be a list of :py:class:`~mlrun.artifacts.Artifact` objects or
            dictionaries, or a single object.
        :return: :py:class:`~mlrun.common.schemas.TagObjects`
        """
        # to avoid circular imports we import here
        import mlrun.artifacts.base

        if not isinstance(artifacts, list):
            artifacts = [artifacts]

        artifact_identifiers = []
        for artifact in artifacts:
            artifact_obj = (
                artifact.to_dict()
                if isinstance(artifact, mlrun.artifacts.base.Artifact)
                else artifact
            )
            artifact_identifiers.append(
                mlrun.common.schemas.ArtifactIdentifier(
                    # we pass the db_key and not the key so the API will be able to find the artifact in the db
                    key=mlrun.utils.get_in_artifact(artifact_obj, "db_key"),
                    uid=mlrun.utils.get_in_artifact(artifact_obj, "uid"),
                    producer_id=mlrun.utils.get_in_artifact(artifact_obj, "tree"),
                    kind=mlrun.utils.get_in_artifact(artifact_obj, "kind"),
                    iter=mlrun.utils.get_in_artifact(artifact_obj, "iter"),
                )
            )
        return mlrun.common.schemas.TagObjects(
            kind="artifact", identifiers=artifact_identifiers
        )

    def get_model_endpoint_monitoring_metrics(
        self,
        project: str,
        endpoint_id: str,
        type: Literal["results", "metrics", "all"] = "all",
    ) -> list[mm_endpoints.ModelEndpointMonitoringMetric]:
        pass

    def get_metrics_by_multiple_endpoints(
        self,
        project: str,
        endpoint_ids: Union[str, list[str]],
        type: Literal["results", "metrics", "all"] = "all",
        events_format: mm_constants.GetEventsFormat = mm_constants.GetEventsFormat.SEPARATION,
    ) -> dict[str, list[mm_endpoints.ModelEndpointMonitoringMetric]]:
        pass

    @abstractmethod
    def delete_project(
        self,
        name: str,
        deletion_strategy: mlrun.common.schemas.DeletionStrategy = mlrun.common.schemas.DeletionStrategy.default(),
    ):
        pass

    @abstractmethod
    def store_project(
        self,
        name: str,
        project: mlrun.common.schemas.Project,
    ) -> mlrun.common.schemas.Project:
        pass

    @abstractmethod
    def patch_project(
        self,
        name: str,
        project: dict,
        patch_mode: mlrun.common.schemas.PatchMode = mlrun.common.schemas.PatchMode.replace,
    ) -> mlrun.common.schemas.Project:
        pass

    @abstractmethod
    def create_project(
        self,
        project: mlrun.common.schemas.Project,
    ) -> mlrun.common.schemas.Project:
        pass

    @abstractmethod
    def list_projects(
        self,
        owner: Optional[str] = None,
        format_: mlrun.common.formatters.ProjectFormat = mlrun.common.formatters.ProjectFormat.name_only,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        state: mlrun.common.schemas.ProjectState = None,
    ) -> mlrun.common.schemas.ProjectsOutput:
        pass

    @abstractmethod
    def get_project(
        self, name: str
    ) -> Union[mlrun.common.schemas.Project, "mlrun.MlrunProject"]:
        pass

    @abstractmethod
    def list_artifact_tags(
        self,
        project=None,
        category: Union[str, mlrun.common.schemas.ArtifactCategories] = None,
    ):
        pass

    @abstractmethod
    def create_feature_set(
        self,
        feature_set: Union[dict, mlrun.common.schemas.FeatureSet],
        project="",
        versioned=True,
    ) -> dict:
        pass

    @abstractmethod
    def get_feature_set(
        self,
        name: str,
        project: str = "",
        tag: Optional[str] = None,
        uid: Optional[str] = None,
    ) -> dict:
        pass

    # TODO: remove in 1.10.0
    @deprecated(
        version="1.7.0",
        reason="'list_features' will be removed in 1.10.0, use 'list_features_v2' instead",
        category=FutureWarning,
    )
    @abstractmethod
    def list_features(
        self,
        project: str,
        name: Optional[str] = None,
        tag: Optional[str] = None,
        entities: Optional[list[str]] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
    ) -> mlrun.common.schemas.FeaturesOutput:
        pass

    @abstractmethod
    def list_features_v2(
        self,
        project: str,
        name: Optional[str] = None,
        tag: Optional[str] = None,
        entities: Optional[list[str]] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
    ) -> mlrun.common.schemas.FeaturesOutputV2:
        pass

    @abstractmethod
    def list_entities_v2(
        self,
        project: str,
        name: Optional[str] = None,
        tag: Optional[str] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
    ) -> mlrun.common.schemas.EntitiesOutputV2:
        pass

    @abstractmethod
    def list_feature_sets(
        self,
        project: str = "",
        name: Optional[str] = None,
        tag: Optional[str] = None,
        state: Optional[str] = None,
        entities: Optional[list[str]] = None,
        features: Optional[list[str]] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        partition_by: Union[
            mlrun.common.schemas.FeatureStorePartitionByField, str
        ] = None,
        rows_per_partition: int = 1,
        partition_sort_by: Union[mlrun.common.schemas.SortField, str] = None,
        partition_order: Union[
            mlrun.common.schemas.OrderType, str
        ] = mlrun.common.schemas.OrderType.desc,
        format_: Union[
            str, mlrun.common.formatters.FeatureSetFormat
        ] = mlrun.common.formatters.FeatureSetFormat.full,
    ) -> list[dict]:
        pass

    @abstractmethod
    def store_feature_set(
        self,
        feature_set: Union[dict, mlrun.common.schemas.FeatureSet],
        name=None,
        project="",
        tag=None,
        uid=None,
        versioned=True,
    ):
        pass

    @abstractmethod
    def patch_feature_set(
        self,
        name,
        feature_set: dict,
        project="",
        tag=None,
        uid=None,
        patch_mode: Union[
            str, mlrun.common.schemas.PatchMode
        ] = mlrun.common.schemas.PatchMode.replace,
    ):
        pass

    @abstractmethod
    def delete_feature_set(self, name, project="", tag=None, uid=None):
        pass

    @abstractmethod
    def create_feature_vector(
        self,
        feature_vector: Union[dict, mlrun.common.schemas.FeatureVector],
        project="",
        versioned=True,
    ) -> dict:
        pass

    @abstractmethod
    def get_feature_vector(
        self,
        name: str,
        project: str = "",
        tag: Optional[str] = None,
        uid: Optional[str] = None,
    ) -> dict:
        pass

    @abstractmethod
    def list_feature_vectors(
        self,
        project: str = "",
        name: Optional[str] = None,
        tag: Optional[str] = None,
        state: Optional[str] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        partition_by: Union[
            mlrun.common.schemas.FeatureStorePartitionByField, str
        ] = None,
        rows_per_partition: int = 1,
        partition_sort_by: Union[mlrun.common.schemas.SortField, str] = None,
        partition_order: Union[
            mlrun.common.schemas.OrderType, str
        ] = mlrun.common.schemas.OrderType.desc,
    ) -> list[dict]:
        pass

    @abstractmethod
    def store_feature_vector(
        self,
        feature_vector: Union[dict, mlrun.common.schemas.FeatureVector],
        name=None,
        project="",
        tag=None,
        uid=None,
        versioned=True,
    ):
        pass

    @abstractmethod
    def patch_feature_vector(
        self,
        name,
        feature_vector_update: dict,
        project="",
        tag=None,
        uid=None,
        patch_mode: Union[
            str, mlrun.common.schemas.PatchMode
        ] = mlrun.common.schemas.PatchMode.replace,
    ):
        pass

    @abstractmethod
    def delete_feature_vector(self, name, project="", tag=None, uid=None):
        pass

    @abstractmethod
    def get_pipeline(
        self,
        run_id: str,
        namespace: Optional[str] = None,
        timeout: int = 30,
        format_: Union[
            str, mlrun.common.formatters.PipelineFormat
        ] = mlrun.common.formatters.PipelineFormat.summary,
        project: Optional[str] = None,
    ):
        pass

    @abstractmethod
    def list_pipelines(
        self,
        project: str,
        namespace: Optional[str] = None,
        sort_by: str = "",
        page_token: str = "",
        filter_: str = "",
        format_: Union[
            str, mlrun.common.formatters.PipelineFormat
        ] = mlrun.common.formatters.PipelineFormat.metadata_only,
        page_size: Optional[int] = None,
    ) -> mlrun.common.schemas.PipelinesOutput:
        pass

    @abstractmethod
    def create_project_secrets(
        self,
        project: str,
        provider: Union[
            str, mlrun.common.schemas.SecretProviderName
        ] = mlrun.common.schemas.SecretProviderName.kubernetes,
        secrets: Optional[dict] = None,
    ):
        pass

    def wait_for_background_task_to_reach_terminal_state(
        self, name: str, project: str = ""
    ) -> mlrun.common.schemas.BackgroundTask:
        pass

    @abstractmethod
    def retry_pipeline(
        self,
        run_id: str,
        project: str,
        namespace: Optional[str] = None,
        timeout: int = 30,
        submit_mode: str = "",
    ):
        pass

    @abstractmethod
    def list_project_secrets(
        self,
        project: str,
        token: str,
        provider: Union[
            str, mlrun.common.schemas.SecretProviderName
        ] = mlrun.common.schemas.SecretProviderName.kubernetes,
        secrets: Optional[list[str]] = None,
    ) -> mlrun.common.schemas.SecretsData:
        pass

    @abstractmethod
    def list_project_secret_keys(
        self,
        project: str,
        provider: Union[
            str, mlrun.common.schemas.SecretProviderName
        ] = mlrun.common.schemas.SecretProviderName.kubernetes,
        token: Optional[str] = None,
    ) -> mlrun.common.schemas.SecretKeysData:
        pass

    @abstractmethod
    def delete_project_secrets(
        self,
        project: str,
        provider: Union[
            str, mlrun.common.schemas.SecretProviderName
        ] = mlrun.common.schemas.SecretProviderName.kubernetes,
        secrets: Optional[list[str]] = None,
    ):
        pass

    @abstractmethod
    def create_user_secrets(
        self,
        user: str,
        provider: Union[
            str, mlrun.common.schemas.SecretProviderName
        ] = mlrun.common.schemas.SecretProviderName.vault,
        secrets: Optional[dict] = None,
    ):
        pass

    @abstractmethod
    def create_model_endpoint(
        self,
        model_endpoint: mlrun.common.schemas.ModelEndpoint,
        creation_strategy: Optional[
            mm_constants.ModelEndpointCreationStrategy
        ] = mm_constants.ModelEndpointCreationStrategy.INPLACE,
    ) -> mlrun.common.schemas.ModelEndpoint:
        pass

    @abstractmethod
    def delete_model_endpoint(
        self,
        name: str,
        project: str,
        function_name: Optional[str] = None,
        function_tag: Optional[str] = None,
        endpoint_id: Optional[str] = None,
    ):
        pass

    @abstractmethod
    def list_model_endpoints(
        self,
        project: str,
        names: Optional[Union[str, list[str]]] = None,
        function_name: Optional[str] = None,
        function_tag: Optional[str] = None,
        model_name: Optional[str] = None,
        model_tag: Optional[str] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        tsdb_metrics: bool = False,
        metric_list: Optional[list[str]] = None,
        top_level: bool = False,
        mode: Optional[mlrun.common.schemas.EndpointMode] = None,
        uids: Optional[list[str]] = None,
        latest_only: bool = False,
    ) -> mlrun.common.schemas.ModelEndpointList:
        pass

    @abstractmethod
    def get_model_endpoint(
        self,
        name: str,
        project: str,
        function_name: Optional[str] = None,
        function_tag: Optional[str] = None,
        endpoint_id: Optional[str] = None,
        tsdb_metrics: bool = True,
        metric_list: Optional[list[str]] = None,
        feature_analysis: bool = False,
    ) -> mlrun.common.schemas.ModelEndpoint:
        pass

    @abstractmethod
    def patch_model_endpoint(
        self,
        name: str,
        project: str,
        attributes: dict,
        function_name: Optional[str] = None,
        function_tag: Optional[str] = None,
        endpoint_id: Optional[str] = None,
    ) -> mlrun.common.schemas.ModelEndpoint:
        pass

    @abstractmethod
    def create_hub_source(
        self, source: Union[dict, mlrun.common.schemas.IndexedHubSource]
    ):
        pass

    @abstractmethod
    def store_hub_source(
        self,
        source_name: str,
        source: Union[dict, mlrun.common.schemas.IndexedHubSource],
    ):
        pass

    @abstractmethod
    def list_hub_sources(
        self,
        item_name: Optional[str] = None,
        tag: Optional[str] = None,
        version: Optional[str] = None,
    ):
        pass

    @abstractmethod
    def get_hub_source(self, source_name: str):
        pass

    @abstractmethod
    def delete_hub_source(self, source_name: str):
        pass

    @abstractmethod
    def get_hub_catalog(
        self,
        source_name: str,
        version: Optional[str] = None,
        tag: Optional[str] = None,
        force_refresh: bool = False,
    ):
        pass

    @abstractmethod
    def get_hub_item(
        self,
        source_name: str,
        item_name: str,
        version: Optional[str] = None,
        tag: str = "latest",
        force_refresh: bool = False,
    ):
        pass

    @abstractmethod
    def verify_authorization(
        self,
        authorization_verification_input: mlrun.common.schemas.AuthorizationVerificationInput,
    ):
        pass

    @abstractmethod
    def store_api_gateway(
        self,
        api_gateway: Union[
            mlrun.common.schemas.APIGateway,
            "mlrun.runtimes.nuclio.api_gateway.APIGateway",
        ],
        project: Optional[str] = None,
    ):
        pass

    @abstractmethod
    def list_api_gateways(self, project=None) -> mlrun.common.schemas.APIGatewaysOutput:
        pass

    @abstractmethod
    def get_api_gateway(self, name, project=None) -> mlrun.common.schemas.APIGateway:
        pass

    @abstractmethod
    def delete_api_gateway(self, name, project=None):
        pass

    @abstractmethod
    def remote_builder(
        self,
        func: "mlrun.runtimes.BaseRuntime",
        with_mlrun: bool,
        mlrun_version_specifier: Optional[str] = None,
        skip_deployed: bool = False,
        builder_env: Optional[dict] = None,
        force_build: bool = False,
    ):
        pass

    @abstractmethod
    def deploy_nuclio_function(
        self,
        func: "mlrun.runtimes.RemoteRuntime",
        builder_env: Optional[dict] = None,
    ):
        pass

    @abstractmethod
    def generate_event(
        self, name: str, event_data: Union[dict, mlrun.common.schemas.Event], project=""
    ):
        pass

    @abstractmethod
    def store_alert_config(
        self,
        alert_name: str,
        alert_data: Union[dict, mlrun.alerts.alert.AlertConfig],
        project="",
        force_reset: bool = False,
    ):
        pass

    @abstractmethod
    def get_alert_config(self, alert_name: str, project=""):
        pass

    @abstractmethod
    def list_alerts_configs(
        self, project="", limit: Optional[int] = None, offset: Optional[int] = None
    ):
        pass

    @abstractmethod
    def delete_alert_config(self, alert_name: str, project=""):
        pass

    @abstractmethod
    def reset_alert_config(self, alert_name: str, project=""):
        pass

    @abstractmethod
    def get_alert_template(self, template_name: str):
        pass

    @abstractmethod
    def list_alert_templates(self):
        pass

    @abstractmethod
    def list_alert_activations(
        self,
        project: Optional[str] = None,
        name: Optional[str] = None,
        since: Optional[datetime.datetime] = None,
        until: Optional[datetime.datetime] = None,
        entity: Optional[str] = None,
        severity: Optional[
            list[Union[mlrun.common.schemas.alert.AlertSeverity, str]]
        ] = None,
        entity_kind: Optional[
            Union[mlrun.common.schemas.alert.EventEntityKind, str]
        ] = None,
        event_kind: Optional[Union[mlrun.common.schemas.alert.EventKind, str]] = None,
    ):
        pass

    @abstractmethod
    def paginated_list_alert_activations(
        self,
        *args,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
        **kwargs,
    ):
        pass

    @abstractmethod
    def get_alert_activation(
        self,
        project,
        activation_id,
    ) -> mlrun.common.schemas.AlertActivation:
        pass

    def update_alert_activation(
        self,
        activation_id: int,
        activation_time: datetime.datetime,
        notifications_states,
    ):
        pass

    @abstractmethod
    def get_builder_status(
        self,
        func: "mlrun.runtimes.BaseRuntime",
        offset: int = 0,
        logs: bool = True,
        last_log_timestamp: float = 0.0,
        verbose: bool = False,
        events_offset: int = 0,
    ):
        pass

    @abstractmethod
    def get_nuclio_deploy_status(
        self,
        func: "mlrun.runtimes.RemoteRuntime",
        last_log_timestamp: float = 0.0,
        verbose: bool = False,
    ):
        pass

    @abstractmethod
    def set_run_notifications(
        self,
        project: str,
        runs: list[mlrun.model.RunObject],
        notifications: list[mlrun.model.Notification],
    ):
        pass

    @abstractmethod
    def store_run_notifications(
        self,
        notification_objects: list[mlrun.model.Notification],
        run_uid: str,
        project: Optional[str] = None,
        mask_params: bool = True,
    ):
        pass

    @abstractmethod
    def get_log_size(self, uid, project=""):
        pass

    @abstractmethod
    def store_alert_notifications(
        self,
        session,
        notification_objects: list[mlrun.model.Notification],
        alert_id: str,
        project: str,
        mask_params: bool = True,
    ):
        pass

    @abstractmethod
    def watch_log(self, uid, project="", watch=True, offset=0):
        pass

    @abstractmethod
    def get_datastore_profile(
        self, name: str, project: str
    ) -> Optional[mlrun.common.schemas.DatastoreProfile]:
        pass

    @abstractmethod
    def delete_datastore_profile(
        self, name: str, project: str
    ) -> mlrun.common.schemas.DatastoreProfile:
        pass

    @abstractmethod
    def list_datastore_profiles(
        self, project: str
    ) -> list[mlrun.common.schemas.DatastoreProfile]:
        pass

    @abstractmethod
    def store_datastore_profile(
        self, profile: mlrun.common.schemas.DatastoreProfile, project: str
    ):
        pass

    @abstractmethod
    def function_status(self, project, name, kind, selector):
        pass

    @abstractmethod
    def start_function(
        self,
        func_url: Optional[str] = None,
        function: "mlrun.runtimes.BaseRuntime" = None,
    ):
        pass

    def get_project_background_task(
        self,
        project: str,
        name: str,
    ) -> mlrun.common.schemas.BackgroundTask:
        pass

    @abstractmethod
    def submit_workflow(
        self,
        project: str,
        name: str,
        workflow_spec: Union[
            "mlrun.projects.pipelines.WorkflowSpec",
            "mlrun.common.schemas.WorkflowSpec",
            dict,
        ],
        arguments: Optional[dict] = None,
        artifact_path: Optional[str] = None,
        source: Optional[str] = None,
        run_name: Optional[str] = None,
        namespace: Optional[str] = None,
        notifications: Optional[list["mlrun.model.Notification"]] = None,
    ) -> "mlrun.common.schemas.WorkflowResponse":
        pass

    @abstractmethod
    def update_model_monitoring_controller(
        self,
        project: str,
        base_period: int = 10,
        image: str = "mlrun/mlrun",
    ) -> None:
        pass

    @abstractmethod
    def enable_model_monitoring(
        self,
        project: str,
        base_period: int = 10,
        image: str = "mlrun/mlrun",
        deploy_histogram_data_drift_app: bool = True,
        fetch_credentials_from_sys_config: bool = False,
    ) -> None:
        pass

    @abstractmethod
    def disable_model_monitoring(
        self,
        project: str,
        delete_resources: bool = True,
        delete_stream_function: bool = False,
        delete_histogram_data_drift_app: bool = True,
        delete_user_applications: bool = False,
        user_application_list: Optional[list[str]] = None,
    ) -> bool:
        pass

    @abstractmethod
    def delete_model_monitoring_function(
        self, project: str, functions: list[str]
    ) -> bool:
        pass

    @abstractmethod
    def set_model_monitoring_credentials(
        self,
        project: str,
        credentials: dict[str, Optional[str]],
        replace_creds: bool,
    ) -> None:
        pass

    @abstractmethod
    def get_monitoring_function_summaries(
        self,
        project: str,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        names: Optional[Union[list[str], str]] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        include_stats: bool = False,
        include_infra: bool = True,
    ) -> list[mlrun.common.schemas.model_monitoring.FunctionSummary]:
        pass

    @abstractmethod
    def get_monitoring_function_summary(
        self,
        project: str,
        function_name: str,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        include_latest_metrics: bool = False,
    ) -> mlrun.common.schemas.model_monitoring.FunctionSummary:
        pass

    @abstractmethod
    def get_project_summary(self, project: str) -> mlrun.common.schemas.ProjectSummary:
        pass

    @abstractmethod
    def get_drift_over_time(
        self,
        project: str,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
    ) -> mlrun.common.schemas.model_monitoring.ModelEndpointDriftValues:
        pass



# Copyright 2025 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
from collections.abc import Awaitable
from typing import Callable, Optional, TypeVar, Union

import mlrun
from mlrun.datastore.model_provider.model_provider import ModelProvider
from mlrun.datastore.utils import accepts_param

T = TypeVar("T")


class OpenAIProvider(ModelProvider):
    """
    OpenAIProvider is a wrapper around the OpenAI SDK that provides an interface
    for interacting with OpenAI's generative AI services.

    It supports both synchronous and asynchronous operations, allowing flexible
    integration into various workflows.

    This class extends the ModelProvider base class and implements OpenAI-specific
    functionality, including client initialization, model invocation, and custom
    operations tailored to the OpenAI API.
    """

    support_async = True

    def __init__(
        self,
        parent,
        schema,
        name,
        endpoint="",
        secrets: Optional[dict] = None,
        default_invoke_kwargs: Optional[dict] = None,
    ):
        endpoint = endpoint or mlrun.mlconf.model_providers.openai_default_model
        if schema != "openai":
            raise mlrun.errors.MLRunInvalidArgumentError(
                "OpenAIProvider supports only 'openai' as the provider kind."
            )
        super().__init__(
            parent=parent,
            kind=schema,
            name=name,
            endpoint=endpoint,
            secrets=secrets,
            default_invoke_kwargs=default_invoke_kwargs,
        )
        self.options = self.get_client_options()
        self.load_client()

    @classmethod
    def parse_endpoint_and_path(cls, endpoint, subpath) -> (str, str):
        if endpoint and subpath:
            endpoint = endpoint + subpath
            #  in openai there is no usage of subpath variable. if the model contains "/", it is part of the model name.
            subpath = ""
        return endpoint, subpath

    def load_client(self) -> None:
        """
        Initializes the OpenAI SDK client using the provided options.

        This method imports the `OpenAI` class from the `openai` package, instantiates
        a client with the given keyword arguments (`self.options`), and assigns it to
        `self._client` and `self._async_client`.

        Raises:
            ImportError: If the `openai` package is not installed.
        """
        try:
            from openai import OpenAI, AsyncOpenAI  # noqa

            self._client = OpenAI(**self.options)
            self._async_client = AsyncOpenAI(**self.options)
        except ImportError as exc:
            raise ImportError("openai package is not installed") from exc

    def get_client_options(self) -> dict:
        res = dict(
            api_key=self._get_secret_or_env("OPENAI_API_KEY"),
            organization=self._get_secret_or_env("OPENAI_ORG_ID"),
            project=self._get_secret_or_env("OPENAI_PROJECT_ID"),
            base_url=self._get_secret_or_env("OPENAI_BASE_URL"),
            timeout=self._get_secret_or_env("OPENAI_TIMEOUT"),
            max_retries=self._get_secret_or_env("OPENAI_MAX_RETRIES"),
        )
        return self._sanitize_options(res)

    def custom_invoke(
        self, operation: Optional[Callable[..., T]] = None, **invoke_kwargs
    ) -> Optional[T]:
        """
        OpenAI-specific implementation of `ModelProvider.custom_invoke`.

        Invokes an OpenAI model operation using the sync client. For full details, see
        `ModelProvider.custom_invoke`.

        Example:
            ```python
            result = openai_model_provider.invoke(
                openai_model_provider.client.images.generate,
                prompt="A futuristic cityscape at sunset",
                n=1,
                size="1024x1024",
            )
            ```
        :param      operation:      Same as ModelProvider.custom_invoke.
        :param      invoke_kwargs:  Same as ModelProvider.custom_invoke.
        :return:                    Same as ModelProvider.custom_invoke.

        """
        invoke_kwargs = self.get_invoke_kwargs(invoke_kwargs)
        model_kwargs = {"model": invoke_kwargs.pop("model", None) or self.model}

        if operation:
            if not callable(operation):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "OpenAI custom_invoke operation must be a callable"
                )
            if not accepts_param(operation, "model"):
                model_kwargs = {}
            return operation(**invoke_kwargs, **model_kwargs)
        else:
            return self.client.chat.completions.create(**invoke_kwargs, **model_kwargs)

    async def async_custom_invoke(
        self,
        operation: Optional[Callable[..., Awaitable[T]]] = None,
        **invoke_kwargs,
    ) -> Optional[T]:
        """
        OpenAI-specific implementation of `ModelProvider.async_custom_invoke`.

        Invokes an OpenAI model operation using the async client. For full details, see
        `ModelProvider.async_custom_invoke`.

        Example:
        ```python
            result = openai_model_provider.invoke(
                openai_model_provider.async_client.images.generate,
                prompt="A futuristic cityscape at sunset",
                n=1,
                size="1024x1024",
            )
        ```

        :param operation:       Same as ModelProvider.async_custom_invoke.
        :param invoke_kwargs:   Same as ModelProvider.async_custom_invoke.
        :return:                Same as ModelProvider.async_custom_invoke.

        """
        invoke_kwargs = self.get_invoke_kwargs(invoke_kwargs)
        model_kwargs = {"model": invoke_kwargs.pop("model", None) or self.model}
        if operation:
            if not inspect.iscoroutinefunction(operation):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "OpenAI async_custom_invoke operation must be a coroutine function"
                )
            if not accepts_param(operation, "model"):
                model_kwargs = {}
            return await operation(**invoke_kwargs, **model_kwargs)
        else:
            return await self.async_client.chat.completions.create(
                **invoke_kwargs, **model_kwargs
            )

    def invoke(
        self,
        messages: Optional[list[dict]] = None,
        as_str: bool = False,
        **invoke_kwargs,
    ) -> Optional[Union[str, T]]:
        """
        OpenAI-specific implementation of `ModelProvider.invoke`.
        Invokes an OpenAI model operation using the sync client.
        For full details, see `ModelProvider.invoke`.

        :param messages:    Same as ModelProvider.invoke.

        :param as_str: bool
                            If `True`, returns only the main content of the first response
                            (`response.choices[0].message.content`).
                            If `False`, returns the full response object, whose type depends on
                            the specific OpenAI SDK operation used (e.g., chat completion, completion, etc.).

        :param invoke_kwargs:
                            Same as ModelProvider.invoke.
        :return:            Same as ModelProvider.invoke.

        """
        response = self.custom_invoke(messages=messages, **invoke_kwargs)
        if as_str:
            return response.choices[0].message.content
        return response

    async def async_invoke(
        self,
        messages: Optional[list[dict]] = None,
        as_str: bool = False,
        **invoke_kwargs,
    ) -> str:
        """
        OpenAI-specific implementation of `ModelProvider.async_invoke`.
        Invokes an OpenAI model operation using the async client.
        For full details, see `ModelProvider.async_invoke`.

        :param messages:    Same as ModelProvider.async_invoke.

        :param as_str: bool
                            If `True`, returns only the main content of the first response
                            (`response.choices[0].message.content`).
                            If `False`, returns the full awaited response object, whose type depends on
                            the specific OpenAI SDK operation used (e.g., chat completion, completion, etc.).

        :param invoke_kwargs:
                            Same as ModelProvider.async_invoke.
        :returns            Same as ModelProvider.async_invoke.

        """
        response = await self.async_custom_invoke(messages=messages, **invoke_kwargs)
        if as_str:
            return response.choices[0].message.content
        return response



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from datetime import timedelta
from typing import Any, Optional, Union

import numpy as np
import storey

import mlrun
import mlrun.artifacts
import mlrun.common.schemas.model_monitoring as mm_schemas
import mlrun.serving
from mlrun.common.schemas import MonitoringData
from mlrun.utils import get_data_from_path, logger


class MonitoringPreProcessor(storey.MapClass):
    """preprocess step, reconstructs the serving output event body to StreamProcessingEvent schema"""

    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.server: mlrun.serving.GraphServer = (
            getattr(self.context, "server", None) if self.context else None
        )

    def reconstruct_request_resp_fields(
        self, event, model: str, model_monitoring_data: dict
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        result_path = model_monitoring_data.get(MonitoringData.RESULT_PATH)
        input_path = model_monitoring_data.get(MonitoringData.INPUT_PATH)

        result = get_data_from_path(result_path, event.body.get(model, event.body))
        output_schema = model_monitoring_data.get(MonitoringData.OUTPUTS)
        input_schema = model_monitoring_data.get(MonitoringData.INPUTS)
        logger.debug("output schema retrieved", output_schema=output_schema)
        if isinstance(result, dict):
            # transpose by key the outputs:
            outputs = self.transpose_by_key(result, output_schema)
            if not output_schema:
                logger.warn(
                    "Output schema was not provided using Project:log_model or by ModelRunnerStep:add_model order "
                    "may not preserved"
                )
        else:
            outputs = result

        event_inputs = event._metadata.get("inputs", {})
        event_inputs = get_data_from_path(input_path, event_inputs)
        if isinstance(event_inputs, dict):
            # transpose by key the inputs:
            inputs = self.transpose_by_key(event_inputs, input_schema)
            if not input_schema:
                logger.warn(
                    "Input schema was not provided using by ModelRunnerStep:add_model, order "
                    "may not preserved"
                )
        else:
            inputs = event_inputs

        if outputs and isinstance(outputs[0], list):
            if output_schema and len(output_schema) != len(outputs[0]):
                logger.info(
                    "The number of outputs returned by the model does not match the number of outputs "
                    "specified in the model endpoint.",
                    model_endpoint=model,
                    output_len=len(outputs[0]),
                    schema_len=len(output_schema),
                )
        elif outputs:
            if output_schema and len(output_schema) != 1:
                logger.info(
                    "The number of outputs returned by the model does not match the number of outputs "
                    "specified in the model endpoint.",
                    model_endpoint=model,
                    output_len=len(outputs),
                    schema_len=len(output_schema),
                )
            if len(inputs) != len(outputs):
                logger.warn(
                    "outputs and inputs are not in the same length check 'input_path' and "
                    "'output_path' was specified if needed"
                )
        request = {"inputs": inputs, "id": getattr(event, "id", None)}
        resp = {"outputs": outputs}

        return request, resp

    @staticmethod
    def transpose_by_key(
        data: dict, schema: Optional[Union[str, list[str]]] = None
    ) -> Union[list[Any], list[list[Any]]]:
        """
        Transpose values from a dictionary by keys.

        Given a dictionary and an optional schema (a key or list of keys), this function:
        - Extracts the values for the specified keys (or all keys if no schema is provided).
        - Ensures the data is represented as a list of rows, then transposes it (i.e., switches rows to columns).
        - Handles edge cases:
            * If a single scalar or single-element list is provided, returns a flat list.
            * If a single key is provided (as a string or a list with one element), handles it properly.
            * If only one row with len of one remains after transposition, unwraps it to avoid nested list-of-one.

        Example::

            transpose_by_key({"a": 1})
            # returns: [1]

            transpose_by_key({"a": [1, 2]})
            # returns: [1 ,2]

            transpose_by_key({"a": [1, 2], "b": [3, 4]})
            # returns: [[1, 3], [2, 4]]

        :param data:     Dictionary with values that are either scalars or lists.
        :param schema:   Optional key or list of keys to extract. If not provided, all keys are used.
                         Can be a string (single key) or a list of strings.

        :return:         Transposed values:
                         * If result is a single column or row, returns a flat list.
                         * If result is a matrix, returns a list of lists.

        :raises ValueError: If the values include a mix of scalars and lists, or if the list lengths do not match.
        """

        # Normalize schema to list
        if not schema:
            keys = list(data.keys())
        elif isinstance(schema, str):
            keys = [schema]
        else:
            keys = schema

        values = [data[key] for key in keys if key in data]
        if len(values) != len(keys):
            raise mlrun.MLRunInvalidArgumentError(
                f"Schema keys {keys} do not match the data keys {list(data.keys())}."
            )

        # Detect if all are scalars ie: int,float,str
        all_scalars = all(not isinstance(v, (list, tuple, np.ndarray)) for v in values)
        all_lists = all(isinstance(v, (list, tuple, np.ndarray)) for v in values)

        if not (all_scalars or all_lists):
            raise ValueError(
                "All values must be either scalars or lists of equal length."
            )

        if all_scalars:
            transposed = np.array([values], dtype=object)
        elif all_lists and len(keys) > 1:
            arrays = [np.array(v, dtype=object) for v in values]
            mat = np.stack(arrays, axis=0)
            transposed = mat.T
        else:
            return values[0]

        if transposed.shape[1] == 1 and transposed.shape[0] == 1:
            # Transform [[0]] -> [0]:
            return transposed[:, 0].tolist()
        return transposed.tolist()

    def do(self, event):
        monitoring_event_list = []
        model_runner_name = event._metadata.get("model_runner_name", "")
        step = self.server.graph.steps[model_runner_name] if self.server else None
        if not step or not hasattr(step, "monitoring_data"):
            raise mlrun.errors.MLRunRuntimeError(
                f"ModelRunnerStep name {model_runner_name} is not found in the graph or does not have monitoring data"
            )
        monitoring_data = step.monitoring_data
        logger.debug(
            "monitoring preprocessor started",
            event=event,
            monitoring_data=monitoring_data,
            metadata=event._metadata,
        )
        if len(monitoring_data) > 1:
            for model in event.body.keys():
                if model in monitoring_data:
                    request, resp = self.reconstruct_request_resp_fields(
                        event, model, monitoring_data[model]
                    )
                    if hasattr(event, "_original_timestamp"):
                        when = event._original_timestamp
                    else:
                        when = event._metadata.get(model, {}).get(
                            mm_schemas.StreamProcessingEvent.WHEN
                        )
                    monitoring_event_list.append(
                        {
                            mm_schemas.StreamProcessingEvent.MODEL: model,
                            mm_schemas.StreamProcessingEvent.MODEL_CLASS: monitoring_data[
                                model
                            ].get(mm_schemas.StreamProcessingEvent.MODEL_CLASS),
                            mm_schemas.StreamProcessingEvent.MICROSEC: event._metadata.get(
                                model, {}
                            ).get(mm_schemas.StreamProcessingEvent.MICROSEC),
                            mm_schemas.StreamProcessingEvent.WHEN: when,
                            mm_schemas.StreamProcessingEvent.ENDPOINT_ID: monitoring_data[
                                model
                            ].get(
                                mlrun.common.schemas.MonitoringData.MODEL_ENDPOINT_UID
                            ),
                            mm_schemas.StreamProcessingEvent.LABELS: monitoring_data[
                                model
                            ].get(mlrun.common.schemas.MonitoringData.OUTPUTS),
                            mm_schemas.StreamProcessingEvent.FUNCTION_URI: self.server.function_uri
                            if self.server
                            else None,
                            mm_schemas.StreamProcessingEvent.REQUEST: request,
                            mm_schemas.StreamProcessingEvent.RESPONSE: resp,
                            mm_schemas.StreamProcessingEvent.ERROR: event.body[model][
                                mm_schemas.StreamProcessingEvent.ERROR
                            ]
                            if mm_schemas.StreamProcessingEvent.ERROR
                            in event.body[model]
                            else None,
                            mm_schemas.StreamProcessingEvent.METRICS: event.body[model][
                                mm_schemas.StreamProcessingEvent.METRICS
                            ]
                            if mm_schemas.StreamProcessingEvent.METRICS
                            in event.body[model]
                            else None,
                        }
                    )
        elif monitoring_data:
            model = list(monitoring_data.keys())[0]
            request, resp = self.reconstruct_request_resp_fields(
                event, model, monitoring_data[model]
            )
            if hasattr(event, "_original_timestamp"):
                when = event._original_timestamp
            else:
                when = event._metadata.get(mm_schemas.StreamProcessingEvent.WHEN)
            monitoring_event_list.append(
                {
                    mm_schemas.StreamProcessingEvent.MODEL: model,
                    mm_schemas.StreamProcessingEvent.MODEL_CLASS: monitoring_data[
                        model
                    ].get(mm_schemas.StreamProcessingEvent.MODEL_CLASS),
                    mm_schemas.StreamProcessingEvent.MICROSEC: event._metadata.get(
                        mm_schemas.StreamProcessingEvent.MICROSEC
                    ),
                    mm_schemas.StreamProcessingEvent.WHEN: when,
                    mm_schemas.StreamProcessingEvent.ENDPOINT_ID: monitoring_data[
                        model
                    ].get(mlrun.common.schemas.MonitoringData.MODEL_ENDPOINT_UID),
                    mm_schemas.StreamProcessingEvent.LABELS: monitoring_data[model].get(
                        mlrun.common.schemas.MonitoringData.OUTPUTS
                    ),
                    mm_schemas.StreamProcessingEvent.FUNCTION_URI: self.server.function_uri
                    if self.server
                    else None,
                    mm_schemas.StreamProcessingEvent.REQUEST: request,
                    mm_schemas.StreamProcessingEvent.RESPONSE: resp,
                    mm_schemas.StreamProcessingEvent.ERROR: event.body[
                        mm_schemas.StreamProcessingEvent.ERROR
                    ]
                    if mm_schemas.StreamProcessingEvent.ERROR in event.body
                    else None,
                    mm_schemas.StreamProcessingEvent.METRICS: event.body[
                        mm_schemas.StreamProcessingEvent.METRICS
                    ]
                    if mm_schemas.StreamProcessingEvent.METRICS in event.body
                    else None,
                }
            )
        event.body = monitoring_event_list
        return event


class BackgroundTaskStatus(storey.MapClass):
    """
    background task status checker, prevent events from pushing to the model monitoring stream target if model endpoints
    creation failed or in progress
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.server: mlrun.serving.GraphServer = (
            getattr(self.context, "server", None) if self.context else None
        )
        self._background_task_check_timestamp = None
        self._background_task_state = mlrun.common.schemas.BackgroundTaskState.running

    def do(self, event):
        if self.server is None:
            return None
        if (
            self._background_task_state
            == mlrun.common.schemas.BackgroundTaskState.running
            and (
                self._background_task_check_timestamp is None
                or mlrun.utils.now_date() - self._background_task_check_timestamp
                >= timedelta(
                    seconds=mlrun.mlconf.model_endpoint_monitoring.model_endpoint_creation_check_period
                )
            )
        ):
            background_task = mlrun.get_run_db().get_project_background_task(
                self.server.project, self.server.model_endpoint_creation_task_name
            )
            self._background_task_check_timestamp = mlrun.utils.now_date()
            self._log_background_task_state(background_task.status.state)
            self._background_task_state = background_task.status.state

        if (
            self._background_task_state
            == mlrun.common.schemas.BackgroundTaskState.succeeded
        ):
            return event
        else:
            return None

    def _log_background_task_state(
        self, background_task_state: mlrun.common.schemas.BackgroundTaskState
    ):
        logger.info(
            "Checking model endpoint creation task status",
            task_name=self.server.model_endpoint_creation_task_name,
        )
        if (
            background_task_state
            in mlrun.common.schemas.BackgroundTaskState.terminal_states()
        ):
            logger.info(
                f"Model endpoint creation task completed with state {background_task_state}"
            )
        else:  # in progress
            logger.info(
                f"Model endpoint creation task is still in progress with the current state: "
                f"{background_task_state}. Events will not be monitored for the next "
                f"{mlrun.mlconf.model_endpoint_monitoring.model_endpoint_creation_check_period} seconds",
                name=self.name,
                background_task_check_timestamp=self._background_task_check_timestamp.isoformat(),
            )


class SamplingStep(storey.MapClass):
    """sampling step, samples the serving outputs for the model monitoring as sampling_percentage defines"""

    def __init__(
        self,
        sampling_percentage: Optional[float] = 100.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.sampling_percentage = (
            sampling_percentage if 0 < sampling_percentage <= 100 else 100
        )

    def do(self, event):
        logger.debug(
            "sampling step runs",
            event=event,
            sampling_percentage=self.sampling_percentage,
        )
        if self.sampling_percentage != 100 and not event.get(
            mm_schemas.StreamProcessingEvent.ERROR
        ):
            request = event[mm_schemas.StreamProcessingEvent.REQUEST]
            num_of_inputs = len(request["inputs"])
            sampled_requests_indices = self._pick_random_requests(
                num_of_inputs, self.sampling_percentage
            )
            if not sampled_requests_indices:
                return None

            event[mm_schemas.StreamProcessingEvent.REQUEST]["inputs"] = [
                request["inputs"][i] for i in sampled_requests_indices
            ]

            if isinstance(
                event[mm_schemas.StreamProcessingEvent.RESPONSE]["outputs"], list
            ):
                event[mm_schemas.StreamProcessingEvent.RESPONSE]["outputs"] = [
                    event[mm_schemas.StreamProcessingEvent.RESPONSE]["outputs"][i]
                    for i in sampled_requests_indices
                ]
        event[mm_schemas.EventFieldType.SAMPLING_PERCENTAGE] = self.sampling_percentage
        event[mm_schemas.EventFieldType.EFFECTIVE_SAMPLE_COUNT] = len(
            event.get(mm_schemas.StreamProcessingEvent.REQUEST, {}).get("inputs", [])
        )
        return event

    @staticmethod
    def _pick_random_requests(num_of_reqs: int, percentage: float) -> list[int]:
        """
        Randomly selects indices of requests to sample based on the given percentage

        :param num_of_reqs: Number of requests to select from
        :param percentage: Sample percentage for each request
        :return: A list containing the indices of the selected requests
        """

        return [
            req for req in range(num_of_reqs) if random.random() < (percentage / 100)
        ]


class MockStreamPusher(storey.MapClass):
    def __init__(self, output_stream=None, **kwargs):
        super().__init__(**kwargs)
        stream = self.context.stream if self.context else None
        self.output_stream = output_stream or stream.output_stream

    def do(self, event):
        self.output_stream.push(
            [event], partition_key=mm_schemas.StreamProcessingEvent.ENDPOINT_ID
        )



# Copyright 2025 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Optional, TypeVar, Union

import mlrun
from mlrun.datastore.model_provider.model_provider import ModelProvider

if TYPE_CHECKING:
    from transformers.pipelines.base import Pipeline

T = TypeVar("T")
ChatType = list[dict[str, str]]  # according to transformers.pipelines.text_generation


class HuggingFaceProvider(ModelProvider):
    """
    HuggingFaceProvider is a wrapper around the Hugging Face Transformers pipeline
    that provides an interface for interacting with a wide range of Hugging Face models.

    It supports synchronous operations, enabling flexible integration into various workflows.

    This class extends the ModelProvider base class and implements Hugging Face-specific
    functionality, including pipeline initialization, default text generation operations,
    and custom operations tailored to the Hugging Face Transformers pipeline API.
    """

    def __init__(
        self,
        parent,
        schema,
        name,
        endpoint="",
        secrets: Optional[dict] = None,
        default_invoke_kwargs: Optional[dict] = None,
    ):
        endpoint = endpoint or mlrun.mlconf.model_providers.huggingface_default_model
        if schema != "huggingface":
            raise mlrun.errors.MLRunInvalidArgumentError(
                "HuggingFaceProvider supports only 'huggingface' as the provider kind."
            )
        super().__init__(
            parent=parent,
            kind=schema,
            name=name,
            endpoint=endpoint,
            secrets=secrets,
            default_invoke_kwargs=default_invoke_kwargs,
        )
        self.options = self.get_client_options()
        self._expected_operation_type = None
        self.load_client()

    @staticmethod
    def _extract_string_output(result) -> str:
        """
        Extracts the first generated string from Hugging Face pipeline output,
        regardless of whether it's plain text-generation or chat-style output.
        """
        if not isinstance(result, list) or len(result) == 0:
            raise ValueError("Empty or invalid pipeline output")

        return result[0].get("generated_text")

    @classmethod
    def parse_endpoint_and_path(cls, endpoint, subpath) -> (str, str):
        if endpoint and subpath:
            endpoint = endpoint + subpath
            # In HuggingFace, "/" in a model name is part of the name — `subpath` is not used.
            subpath = ""
        return endpoint, subpath

    def load_client(self) -> None:
        """
        Initializes the Hugging Face pipeline using the provided options.

        This method imports the `pipeline` function from the `transformers` package,
        creates a pipeline instance with the specified task and model (from `self.options`),
        and assigns it to `self._client`.

        Note: Hugging Face pipelines are synchronous and do not support async invocation.

        Raises:
            ImportError: If the `transformers` package is not installed.
        """
        try:
            from transformers import pipeline, AutoModelForCausalLM  # noqa
            from transformers import AutoTokenizer  # noqa
            from transformers.pipelines.base import Pipeline  # noqa

            self._client = pipeline(model=self.model, **self.options)
            self._expected_operation_type = Pipeline
        except ImportError as exc:
            raise ImportError("transformers package is not installed") from exc

    def get_client_options(self):
        res = dict(
            task=self._get_secret_or_env("HF_TASK") or "text-generation",
            token=self._get_secret_or_env("HF_TOKEN"),
            device=self._get_secret_or_env("HF_DEVICE"),
            device_map=self._get_secret_or_env("HF_DEVICE_MAP"),
            trust_remote_code=self._get_secret_or_env("HF_TRUST_REMOTE_CODE"),
            model_kwargs=self._get_secret_or_env("HF_MODEL_KWARGS"),
        )
        return self._sanitize_options(res)

    def custom_invoke(
        self, operation: Optional["Pipeline"] = None, **invoke_kwargs
    ) -> Optional[T]:
        """
        HuggingFace implementation of `ModelProvider.custom_invoke`.
        Use the default config in provider client/ user defined client:

        Example:
        ```python
            image = Image.open(image_path)
            pipeline_object =  pipeline("image-classification", model="microsoft/resnet-50")
            result = hf_provider.custom_invoke(
                pipeline_object,
                inputs=image,
            )
        ```


        :param operation:               A pipeline object
        :param invoke_kwargs:           Keyword arguments to pass to the operation.
        :return:                        The full response returned by the operation.

        """
        invoke_kwargs = self.get_invoke_kwargs(invoke_kwargs)
        if operation:
            if not isinstance(operation, self._expected_operation_type):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "Huggingface operation must inherit" " from 'Pipeline' object"
                )
            return operation(**invoke_kwargs)
        else:
            return self.client(**invoke_kwargs)

    def invoke(
        self,
        messages: Union[str, list[str], ChatType, list[ChatType]] = None,
        as_str: bool = False,
        **invoke_kwargs,
    ) -> Optional[Union[str, list, T]]:
        """
        HuggingFace-specific implementation of `ModelProvider.invoke`.
        Invokes a HuggingFace model operation using the synchronous client.
        For complete usage details, refer to `ModelProvider.invoke`.
        :param messages:
                            Same as ModelProvider.invoke.

        :param as_str:
                            If `True`, returns only the main content from a single response
                            (intended for single-response use cases).
                            If `False`, returns the full response object, whose type depends on
                            the client (e.g., `pipeline`).

        :param invoke_kwargs:
                            Same as ModelProvider.invoke.
        :return:            Same as ModelProvider.invoke.
        """
        if self.client.task != "text-generation":
            raise mlrun.errors.MLRunInvalidArgumentError(
                "HuggingFaceProvider.invoke supports text-generation task only"
            )
        if as_str:
            invoke_kwargs["return_full_text"] = False
        response = self.custom_invoke(text_inputs=messages, **invoke_kwargs)
        if as_str:
            return self._extract_string_output(response)
        return response
